/* $Id: inode-cache.c,v 1.2 2000/10/10 22:10:44 dps Exp $ */
/* Inode cache */

#include "config.h"
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif /* HAVE_STDIO_H */
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif /* HAVE_STDLIB_H */
#ifdef HAVE_STRING_H
#include <string.h>
#endif /* HAVE_STRING_H */
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif /* HAVE_STRINGS_H */
#ifdef HAVE_TIME_H
#include <time.h>
#endif /* HAVE_TIME_H */
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif /* HAVE_SYS_TIME_H */
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif /* HAVE_CTYPE_H */
#include "detail.h"

#define ICACHE_HASH 2017	/* Hash table size */
#define HPRIME 65537		/* Big prime */
#ifndef ICACHE_MAX
#define ICACHE_MAX 10000	/* MAx hash table entries */
#endif /* ICACHE_MAX */

typedef struct icache_entry
{
    const char *fname;
    ino_t inode;
    dev_t dev;
    time_t mtime;
    struct icache_entry *next;	   /* Nect with same hash value */
    struct icache_entry **prevp;   /* Previous with same hash value */
    struct icache_entry *lru;	   /* Next in LRU chain */
    struct icache_entry **lrup;	   /* Pointer to me in LRU chain */
} icache_entry;

static icache_entry *htab[ICACHE_HASH];
static int icache_entries=0;	/* #entries */

static icache_entry *lru=NULL;	/* Head of lru chain */
static icache_entry **lru_top=&lru; /* Last item on lru chain */

/* Initialise hash table */
void icache_init(void)
{
    int i;
    for (i=0; i<ICACHE_HASH; i++)
	htab[i]=NULL;
}

static char *trim(char *s)
{
    char *t;
    t=s+strlen(s)-1;
    while (t>=s)
    {
	if (!isspace(*t))
	    break;
	*(t--)='\0';
    }
    return s;
}

/* Make a new node */
static icache_entry *make_entry(const char *fname, ino_t inode, dev_t dev, time_t mt)
{
    icache_entry *res;

#ifdef DEBUG
    fprintf(stderr, "Icache: adding%s\n", fname);
#endif /* DEBUG */
    if ((res=(icache_entry *) malloc(sizeof(icache_entry)))==NULL)
	return NULL;
    if ((res->fname=strdup(fname))==NULL)
    {
	free(res);
	return NULL;
    }
    res->inode=inode;				/* Inode */
    res->dev=dev;				/* Device */
    res->mtime=mt;				/* Modification time */
#ifdef DEBUG
    fprintf(stderr, "Icache: added %s\n", fname);
#endif /* DEBUG */
    return res;
}

/* The hash function */
static int hash(const char *name)
{
    int hval=0;
    while (*name)
    {
	hval=hval*HPRIME+(*name);
	name++;
    }
    return hval % ICACHE_HASH;
} 

/* If too many entries */
static void remove_lru(void)
{
    icache_entry *p;
#ifdef DEBUG
    fprintf(stderr, "Icache: expiring %s\n", lru->fname);
#endif /* DEBUG */
    p=lru;			/* Save lru */
    lru=lru->lru;		/* Update lru */
    lru->lrup=&lru;		/* lru points to this entry now */
    *(p->prevp)=p->next;	/* Bypass p */
    if (p->next!=NULL)
    {
	p->next->prevp=p->prevp; /* Update scan->next's prevp */
	free((void *) p->fname); /* Free name */
	free(p);
    
    }
#ifdef DBUG
    fprintf(stderr, "Icache: expired %s\n", lru->fname);
#endif /* DEBUG */
} 

static void top_lru(icache_entry *node)
{
    if (node->lru!=NULL)
    {
	*(node->lrup)=node->lru;	/* Snip out scan */
	node->lru->lrup=node->lrup;	/* scan->lru known non-NULL */
	*lru_top=node;			/* Point lru_top to scan */
	node->lru=NULL;			/* No next item in LRU queue */
	node->lrup=lru_top;		/* Pointer at lru_top */
	lru_top=&(node->lru);		/* lru_top is scan->lru */
    }
}

/* Check file is same as we saw before */
void inode_cache(const char *fname, ino_t inode, dev_t dev, time_t mtime, int fixed)
{
    icache_entry *scan, **fix;
    int hv;
    char *tim, *tim2;

    hv=hash(fname);
    scan=htab[hv];
    fix=&(htab[hv]);
#ifdef DEBUG
    fprintf(stderr, "Icache: %s hashes to %d (%p), total cache %d\n",
	    fname, hv,(void *)  htab[hv], icache_entries);
#endif
    while(scan!=NULL)
    {
#ifdef DEBUG
	fprintf(stderr, "Item %s (ino=%d dev=%d), want %s (ino=%d dev=%d)\n",
		scan->fname, (int) scan->inode, (int) scan->dev,
		fname, (int) inode, (int) dev);
#endif /* DEBUG */
	if (strcmp(fname, scan->fname)==0)
	{
	    if (inode!=scan->inode || dev!=scan->dev)
	    {
		log("File %s changed from inode %d dev %d to inode %d dev %d",
		    fname, (int) scan->inode, (int) scan->dev,
		    (int) inode, (int) dev);
		/* Prevent more warnings */
		scan->inode=inode;	/* Update inode */
		scan->dev=dev;		/* Update dev */
		scan->mtime=mtime;	/* Update mtime */
	    } else if (mtime!=scan->mtime && (fixed & INODE_MTIME)!=0)
	    {
		tim=strdup(ctime(&mtime));		    /* Convert mtime */
		tim2=strdup(ctime(&(scan->mtime)));	    /* Convert scan->mtime */
		trim(tim);			    /* My ctime adds \n */
		trim(tim2);			    /* My ctime adds \n */
		log("File %s: mtime changed from %ul (%s) to %ul (%s)",
		    fname,
		    (unsigned long) mtime,
		    (tim==NULL) ? "<strdup failure>" : tim,
		    (unsigned long) scan->mtime,
		    (tim2==NULL) ? "<strdup failure>" : tim2);
		scan->mtime=mtime;
		free(tim);
		free(tim2);
	    }
#ifdef ICACHE_MAX
	    /* Make sure scan is last on LRU queuue */
	    top_lru(scan);

#endif /* ICACHE_MAX */
	    return;
	}
	fix=&(scan->next);
	scan=scan->next;
    }

    /* Allocate new entry */
    if ((scan=make_entry(fname, inode, dev, mtime))==NULL)
    {
	log("icache: malloc failure (this *might* be innocent but check)");
	return;
    }

    scan->next=NULL;		/* No next */
    scan->prevp=fix;		/* Pointed at by fix */
    *fix=scan;			/* Link in scan */
#ifdef DEBUG
    fprintf(stderr, "Icache: now htab[%d]=%p\n", hv, htab[hv]);
#endif /* DEBUG */
    icache_entries++;		/* Add an icache entry  */
    if (fixed & INODE_KEEP)
    {
	scan->lru=NULL;		/* Random values will cause bugs */
	return;			/* If fixed, do not link to lru  */
    }

    /* Link item to lru list */
    scan->lru=NULL;		/* No next in LRU queue */
    scan->lrup=lru_top;		/* Pointed at by lru_top */
    *lru_top=scan;		/* Scan is last of LRU queue */
    lru_top=&(scan->lru);	/* Update lru_top */

    if (icache_entries>ICACHE_MAX)
    {
	remove_lru();		/* Remover leats recently used item */
	icache_entries--;	/* Update cache size  */
    }
}
