/* $Id: safeexec.c,v 1.2 2000/10/10 22:10:44 dps Exp $ */
#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif /* HAVE_CONFIG_H */
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif /* HAVE_STDLIB_H */
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif /* HAVE_SIGNAL_H */
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif /* HAVE_FCNTL_H */
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif /* HAVE_SYS_STAT_H */
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif /* HAVE_SYS_TYPES_H */
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif /* HAVE_TIME_H */
#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif /* HAVE_SYS_RESOURCE_H */
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif /* HAVE_ERRNO_H */
#include "prog.h"
#include "safeexec_p.h"
#include "cfg_check.h"
#include "detail.h"

static int fix_setid(const struct stat *stbuf, const char *fname)
{
    struct stat stbuf2;
    FILE *f;
    int ret, do_chmod=0;
    
    if ((f=fopen(fname, "r"))==NULL)
    {
	log("Could not open %s", fname);
	exit(0);
    }
    if (fstat(fileno(f), &stbuf2))
    {
	fclose(f);
	log("Stat %s as openned file failed (error %s)", fname,
	    strerror(errno));
	exit(0);
    }
    
#ifdef PARANIOD
    /* Ultra-paranoia check */
    if (stbuf->st_dev!=stbuf2.st_dev || stbuf->st_ino!=stbuf2.st_ino)
    {
	log (quote(PS_PATH) " had dev/ino=%d/%d file had %d/%d",
	     stbuf->st_dev, stbuf->st_ino, stbuf2.st_dev, stbuf2.st_ino);
	exit(0);
    }
#endif /* PARANIOD */

    if (!S_ISREG(stbuf2.st_mode) || !S_ISREG(stbuf2.st_mode))
    {
	log("%s is not a regular file", fname);
	fclose(f);
	exit(0);
    }
    if (stbuf2.st_mode & SUID_MASK)
    {
	log("Killing SUID bit of %s", fname);
	do_chmod=1;
    }
    if (stbuf2.st_mode & SGID_MASK)
    {
	log("Killing SGID bit of %s", fname);
	do_chmod=1;
    }

    if (do_chmod)
#ifdef HAVE_FCHMOD
	ret=fchmod(fileno(f), stbuf2.st_mode & ~(SGID_MASK | SUID_MASK));
#else
        ret=chmod(fname, stbuf2.st_mode & ~(SGID_MASK | SUID_MASK));
#endif /* not HAVE_FCHMOD */
    else
	ret=0;
    fclose(f);
    return ret;
}

/* Exec with built-in paranoia */
int safe_exec(const char *fname, const char *const *args, int fd)
{
#if defined(HAVE_SETRLIMIT) && !defined(TEST)
#ifdef HAVE_RLIMIT_NPROC
    static const struct rlimit pid_lim={RLIMIT_NPROC, MAX_PROCS};
#endif /* HAVE RLIMIT_NPROC */
#ifdef HAVE_RLIMIT_MEMLOCK
    static const struct rlimit mlock_lim={RLIMIT_MEMLOCK, 0};
#endif /* HAVE_RLIMIT_MEMLOCK */
#endif /* HAVE_SETRLIMIT and !TEST */
    struct stat stbuf, stbuf3;
    gid_t gid;
    int null;			/* Fd of /dev/null */

    signal(SIGTSTP, SIG_IGN); /* Stop TSTP */
    if ((null=open("/dev/null", O_RDWR))==-1)
    {
       log("Could not open /dev/null (%s)", strerror(errno));
       return 0;
    }

#ifdef DEV_NULL_RDEV
    if (fstat(null, &stbuf))
    {
	close(null);
	log("fstat of fd %d (=/dev/null) failed %s",
	    strerror(errno));
	exit(0);
    }
    if (stbuf.st_rdev!=DEV_NULL_RDEV)
    {
	log("/dev/null has st_rdev %04x, should be %04x",
	    (int) stbuf.st_rdev, (int) DEV_NULL_RDEV);
	close(null);
	exit(0);
    }
#endif /* DEV_NULL_RDEV */

    /* I know this is not race condition proof but worth soemthing
     * anyway (should make life harder for crackers...) */
    if (lstat(fname, &stbuf))
    {
	log("Stat %s  failed\n", fname);
	close(null);
	exit(0);
    }
    
    if (fix_setid(&stbuf, fname)!=0)
    {
	log("Failed to clear setid bis of %s (error %s)", fname,
	    strerror(errno));
    }

    /* Now check inode cache */
    inode_cache(fname, stbuf.st_ino, stbuf.st_dev, stbuf.st_mtime,
		INODE_KEEP | INODE_MTIME);
    

#if defined(HAVE_SETRLIMIT) && !defined(TEST)
#ifdef HAVE_RLIMIT_NPROC
    setrlimit(RLIMIT_NPROC, (struct rlimit *) &pid_lim);     /* set procs limit */
#endif /* HAVE_RLIMIT_NPROC */
#ifdef HAVE_RLIMIT_MEMLOCK
    setrlimit(RLIMIT_MEMLOCK, (struct rlimit *) &mlock_lim); /* set mlock limit */
#endif /* HAVE_RLMIT_MEMLOCK */
#endif /* HAVE_SETRLIMIT && !TEST */

    /* Fix fd's */
    close(0);
    close(1);
    dup2(null,0);
    dup2(fd,1);
    /* Set on exec for all other fds */
    for (fd=3; fd<1024; fd++)	/* FIXME: Magic number */
	fcntl(fd, F_SETFD);	/* Set close-on-exec flag  */

    /* Drop privs */
    gid=DEFGID;
#ifndef TEST
    if (become_gid(DEFGID)!=0 || setgroups(1, &gid)!=0)
    {
	log("Error %s droping group privs, not executing %s",
	    strerror(errno), fname);
	close(null);
	exit(0);
    }
    if (become_uid(DEFUID)!=0)
    {
	log("Error %s droping user privs, not executing %s",
	    strerror(errno), fname);
	close(null);
	exit(0);
    }
#endif /* not TEST */

#ifndef DEBUG
    close(2);
    dup2(null,2);
#endif
         
#ifdef PARANOID
   /* Final paranioa check... */
    if (stat(quote(PS_PATH), &stbuf3))
    {
	log("%s: stat failed on third try", fname);
	exit(0);
    }
    if (stbuf3.st_mode!=stbuf.st_mode & ~(SUID_MASK | SGID_MASK)))
    {
	log("Permission changed since I opened %s". fname);
	exit(0);
    }
    
    /* We know that stbuf2.st_dev==stbuf.st_dev and
     * stbuf2.st_ino==stbuf.st_ino here. */
    if (stbuf3.st_dev!=stbuf.st_dev || stbuf3.st_ino!=stbuf.st_ino)
    {
	log ("%s changed from dev/ino=%d/%d to %d/%d", fname,
	     stbuf.st_dev, stbuf.st_ino, stbuf3.st_dev, stbuf3.st_ino);
	exit(0);
    }
#endif /* PARANIOD */

    /* Run program */
    execv(fname, (char *const *) args);
    exit(0);			 /* Should never get here */
}   
