/* $Id$ */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif /* HAVE_SIGNAL_H */
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif /* HAVE_SYS_WAIT_H */
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif /* HAVE_SYS_SOCKET_H */
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif /* HAVE_NETINET_IN_H */
#ifdef HAVE_SYS_UN_H
#include <sys/un.h>
#endif /* HAVE_SYS_UN_H */
#ifdef HAVE_SETJMP_H
#include <setjmp.h>
#endif /* HAVE_SETJMP_H */
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */

#include "libdetail/detail.h"
#include "socklist.h"

#ifndef N
#define N(x) (sizeof(x)/sizeof(x[0]))
#endif /* N */

#ifndef PATHMAX
#define PATHMAX 2048
#endif /* PATHMAX */

#ifndef NSTAT_MAXTIME
#define NSTAT_MAXTIME 300
#endif /* NSTAT_MAXTIME */

#ifndef quote
#define __quote(x) #x
#define quote(x) __quote(x)
#endif /* quote */

/* Unimplemented type or bungled netstat lies */
static socklist *unimp_rd(const char *s, FILE *in, socklist *sock)
{
    in=in;			/* Stop compiler warning */
    sock=sock;			/* Ditto */
    s=s;			/* Ditto */
    return NULL;
}


/* Read tcp detials, toilet roll style */
static socklist  *tcp_read(const char *s, FILE *in, socklist *sock)
{
    struct sockaddr_in *src, *dst;
    static const char *const tcp_states[]=
    {
	"LISTEN", "ESTABLISHED", "SYN_SENT", "SYN_RECV", "FIN_WAIT1", "FIN_WAIT2",
	"TIME_WAIT", "CLOSED", "CLOSE_WAIT", "LAST_ACK", "CLOSING", "UNKNOWN"
    };
    int i, user, digits;
    uid_t uid;

    if (!check_fmt(in, " %d %d "))
	return NULL;		/* Verify Send-Q and Recv-Q reasonable */

    /* Read local address */
    if ((src=(struct sockaddr_in *) malloc(sizeof(struct sockaddr_in)))==NULL)
	return NULL;

    if (!read_ip(in, src))
	goto out1;

    absorb_ws(in);		/* Eat seperating white space */

    /* Read remote address */
    if ((dst=(struct sockaddr_in *) malloc(sizeof(struct sockaddr_in)))==NULL)
	goto out1;		/* Free src and exit */

    if (!read_ip(in, dst))
	goto out;		/* Free src and exit */

    if (absorb_ws(in)<=0)
	goto out;		/* The code police will get me */

    /* Check the state is legitimate */
    if (!check_list(in, tcp_states, N(tcp_states)))
	goto out;

    /* Read the user id and check it is legit */
    if (absorb_ws(in)<=0)
	goto out;
    if ((user=readdec(in, &digits))==-1)
	goto out;
    uid=(uid_t) user;
    if (user!=(int) uid)
    {
	log("Impossible uid %d (int uid_t %d)", user, (int) uid);
	goto out;
    }

    /* Check end of line */
    if (!check_eol(in))
	goto out;

    /* Everything OK, fill in the detials */
    sock->src=(struct sockaddr *) src;
    sock->dest=(struct sockaddr *) dst;
    sock->dev=UNKNOWN_DEVICE;	/* Dev not unknown */
    sock->inode=UNKNOWN_INODE;	/* Inode not known */
    sock->uid=uid;		/* User id */
    sock->refs=NULL;		/* Can not do refs here */
    return sock;		/* Return succesfully */

 out:
    free(dst);
 out1:
    free(src);
    return NULL;
}


/* Read udp detials, toilet roll style */
static socklist  *udp_read(const char *s, FILE *in, socklist *sock)
{
    struct sockaddr_in *src, *dst;

    if (!check_fmt(in, " %d %d "))
	return NULL;		/* Verify Send-Q and Recv-Q reasonable */

    /* Read local address */
    if ((src=(struct sockaddr_in *) malloc(sizeof(struct sockaddr_in)))==NULL)
	return NULL;

    if (!read_ip(in, src))
	goto out1;

    absorb_ws(in);		/* Eat seperating white space */

    /* Read remote address */
    if ((dst=(struct sockaddr_in *) malloc(sizeof(struct sockaddr_in)))==NULL)
	goto out1;		/* Free src and exit */

    if (!read_ip(in, dst))
	goto out;		/* Free src and exit */

    /* Check end of line */
    if (!check_eol(in))
	goto out;

    /* Everything OK, fill in the detials */
    sock->src=(struct sockaddr *) src;
    sock->dest=(struct sockaddr *) dst;
    sock->dev=UNKNOWN_DEVICE;	/* Dev not unknown */
    sock->inode=UNKNOWN_INODE;	/* Inode not known */
    sock->uid=UNKNOWN_UID;	/* User id not known */
    sock->refs=NULL;		/* Can not do refs here */
    return sock;		/* Return succesfully */

 out:
    free(dst);
 out1:
    free(src);
    return NULL;
}


/* Read raw detials */
static socklist  *raw_read(const char *s, FILE *in, socklist *sock)
{
    struct sockaddr_in *src, *dst;

    if (!check_fmt(in, " %d %d "))
	return NULL;		/* Verify Send-Q and Recv-Q reasonable */

    /* Read local address */
    if ((src=(struct sockaddr_in *) malloc(sizeof(struct sockaddr_in)))==NULL)
	return NULL;

    if (!read_ip(in, src))
	goto out1;

    absorb_ws(in);		/* Eat seperating white space */

    /* Read remote address */
    if ((dst=(struct sockaddr_in *) malloc(sizeof(struct sockaddr_in)))==NULL)
	goto out1;		/* Free src and exit */

    if (!read_ip(in, dst))
	goto out;		/* Free src and exit */

    /* Check end of line */
    if (!check_eol(in))
	goto out;

    /* Everything OK, fill in the detials */
    sock->src=(struct sockaddr *) src;
    sock->dest=(struct sockaddr *) dst;
    sock->dev=UNKNOWN_DEVICE;	/* Dev not unknown */
    sock->inode=UNKNOWN_INODE;	/* Inode not known */
    sock->uid=UNKNOWN_UID;	/* User id not known */
    sock->refs=NULL;		/* Can not do refs here */
    return sock;		/* Return succesfully */

 out:
    free(dst);
 out1:
    free(src);
    return NULL;
}


static const char *const unix_states[]=
{ "LISTENING", "CONNECTED", "CONNECTING", "DISCONNECTED",
  "FREE", "UNKNOWN" };

/* Test for inconsistent flags, state and links for unix stream socket */
static void unix_sock_state(int flg, int nlinks, int state, const char *path)
{
    if ((flg & 8)==8)				/* ACC */
    {
	if (nlinks!=0 || state!=1)
	{
	    log("Suspect socket: ACC and links %d, state %s, path %s",
		nlinks, unix_states[state-1], path);
	}
    }
    else
    {
	if (nlinks==0 || state==1)
	{
	    log("Suspect socket: non-ACC and links %d, state %s, path %s",
		nlinks, unix_states[state-1], path);
	}
    }
}
    
/* Read a unix socket information */
static socklist *unix_read(const char *s, FILE *in, socklist *sock)
{
    static struct
    {
	const  char *name;
	enum __socket_type type;
    } unix_types[]=
    {
	{ "STREAM", SOCK_STREAM }, { "DGRAM", SOCK_DGRAM },
	{ "RAW", SOCK_RAW }, { "RDM", SOCK_RDM },
	{ "SEQPACKET", SOCK_SEQPACKET }
    };
    static const char *const flags[]=
    { "ACC ", "W ", "N " };				/* from netstat.c */
    int flg, state, proto;				/* Flags set */
    int ino, nlinks, c, i;				/* Interesting nos */
    int digits;						/* Dummy */
    ino_t inode;					/* Possible inode # check */
    char path[PATHMAX];					/* Path */
    struct sockaddr_un *src;				/* Address */
    
    if (absorb_ws(in)<=0)
	return NULL;
    if ((nlinks=readdec(in, &digits))==-1)
	return NULL;

    /* Check the flags */
    if (absorb_ws(in)<=0)
	return NULL;
    if (!require_string(in, "[ "))
	return NULL;
    for (flg=0, i=0; i<N(flags); i++)
    {
	if (get_string(in, path, 4)==-1)
	    return NULL;
	if ((c=fgetc(in))!=' ')
	    return NULL;
	while(i<N(flags))
	{
	    if (strcmp(flags[i], path)==0)
	    {
		flg +=1;
		break;
	    }
	    i++;
	    flg <<=1;
	}
    }
    if (strcmp(path, "]")!=0)
	return NULL;
    
    /* Check the type */
    if(absorb_ws(in)<=0)
	return NULL;
    if (get_string(in, path, 10)==-1)
	return NULL;
    for (proto=-1, i=0; i<N(unix_types); i++)
    {
	if (strcmp(unix_types[i].name, path)==0)
	{
	    proto=i;
	    break;
	}
    }
    if (proto==-1)
	return NULL;

    /* Check state */
    if (absorb_ws(in)<=0)
	return NULL;
    if ((state=check_list(in, unix_states, N(unix_states)))==0)
	return NULL;
    
    /* Read inode */
    if (absorb_ws(in)<0)
	return NULL;
    if ((ino=readdec(in, &digits))==-1)
	return NULL;

    /* Path */
    if (absorb_ws(in)<0)
	return NULL;
    
    for (i=0; i<sizeof(path); i++)
    {
	if ((c=fgetc(in))==EOF)
	    break;
	if (c=='\n')
	    break;
	path[i]=c;
    }
    if (i==sizeof(path))
    {
	log("Warning: inode %d, incomplete path `%s...'", ino, s);
	i--;
    }
    path[i]='\0';
    
    switch(proto)
    {
    case 0:						/* STREAM */
	unix_sock_state(flg, nlinks, state, path);		/* Paranioa check */
	break;
    default:
	break;
    }
    inode=(ino_t) ino;
    if (ino!=(int) inode)
    {
	log("Impossible inode number %d (as ino_t is %d) for path %s",
	    ino, (int) inode, s);
    }
    
    if ((src=(struct sockaddr_un *) malloc(sizeof(struct sockaddr_un)))==NULL)
    {
	log("unix_read: malloc failure links %d, proto %s, state %s, inode %d, path %s",
	    nlinks, unix_types[proto].name, unix_states[state-1], (int) inode, s);
	return NULL;
    }
    
    src->sun_family=AF_UNIX;
    strncpy(src->sun_path, path, sizeof(src->sun_path));
    *(src->sun_path+sizeof(src->sun_path)-1)='\0';
    sock->type=unix_types[proto].type;			/* Afjust type */
    sock->src=(struct sockaddr *) src;			/* Source address */
    sock->dest=NULL;					/* Destination n/a */
    sock->dev=UNKNOWN_DEVICE;				/* Unknown device */
    sock->inode=inode;					/* Inode */
    sock->uid=UNKNOWN_UID;				/* Uknown uid */
    sock->refs=NULL;					/* Refs not known */
    return sock;
}


/* Read one line of netstat */
static socklist *process_line(FILE *in)
{
    static const struct
    {
	const char *name;	 /* Name in netstat */
	enum __socket_type type; /* Type */
	sa_family_t fam;	 /* Family or (sa_family_t) -1 */
	socklist *(*rdsock)(const char *, FILE *, socklist *);	
    } socks[]=
    {
	{ "tcp", SOCK_STREAM, AF_INET, tcp_read },	/* tcp */
	{ "udp", SOCK_DGRAM, AF_INET, udp_read  },	/* udp */
	{ "rdp", SOCK_SEQPACKET, AF_INET, unimp_rd },	/* rfc 1151 */
	{ "raw", SOCK_RAW, AF_INET, raw_read },		/* raw sockets */
	{ "unix", SOCK_STREAM, AF_UNIX, unix_read}	/* unix domain sockets */
    };
    char s[20];				    /* More than required */
    int i;
    socklist *sock, *nsock;

    switch(get_string(in, s, sizeof(s)))
    {
    case -1:
	log("WARNING: Initial string >%d. Killing.", sizeof(s));
	kill(spawned_pid(), SIGKILL);		     /* Nuke bad netstat */
	break;

    case 0:
	absorb_ws(in);
	break;

    case 5:
	if (strcmp(s, "Proto")==0)
	{
	    eat_line(in);
	    break;
	}
	eat_line(in);
	break;

    case 6:
	if (strcmp(s, "Active")==0)
	{
	    eat_line(in);
	    break;
	}
	eat_line(in);
	break;

    default:
	for (i=0; i<N(socks); i++)
	{
	    if (strcmp(s, socks[i].name)==0)
	    {
		sock=(socklist *) malloc(sizeof(socklist));
		if (sock==NULL)
		{
		    eat_line(in);
		    break;
		}
		sock->fam=socks[i].fam;
		sock->type=socks[i].type;
		nsock=(*socks[i].rdsock)(s, in, sock);
		if (nsock==NULL)
		{
		    log("Netstat output format violated, suspect cracking");
		    free(sock);
		    eat_line(in);
		}
		return nsock;
		/* NOT REACHED */
	    }
	}
	eat_line(in);
	break;
    }
    return NULL;
}


/* Run netstat */
socklist *run_netstat(void)
{
    FILE *in;
    socklist *sock, *res, **fix;
    int c, num, nc, i;
    static const char *const nstat_args[]={"netstat", "-ane", NULL}; 
 
    res=NULL;						/* Empty list */
    fix=&res;						/* Change res */

    if ((in=safe_spawn(quote(PATH_NETSTAT), nstat_args, NSTAT_MAXTIME))==NULL)
	return NULL;

    while ((c=fgetc(in))!=EOF)
    {
	ungetc(c, in);
	sock=process_line(in);
	if (sock!=NULL)
	{
	    /* Add socket to list */
	    sock->next=NULL;
	    *fix=sock;
	    fix=&(sock->next);
	}
    }
    return res;
}



/******************** Testing ********************/	

#ifdef TEST
int main(void)
{
    socklist *socks;

    icache_init();

    socks=run_netstat();
    print_socklist(socks);
    socklist_zap(socks);
    return 0;
}
#endif /* TEST */
